/*
  Description  : Use the class to format the number string with localize decimal and
                 thousand separator charactors.
  Developer    : Simen
  Contact      : SimenWu@hotmail.com
  Last Modified: 2/22/2003
*/

#include "stdafx.h"
#include "LocaleEngine.h"

LocaleEngine::LocaleEngine() 
{
  buf_size=20;
  numFormat.lpDecimalSep =new char[buf_size];
  numFormat.lpThousandSep=new char[buf_size];
  Init();
}

LocaleEngine::~LocaleEngine()
{
  delete [](numFormat.lpDecimalSep);
  delete [](numFormat.lpThousandSep);
}

void LocaleEngine::Init()
{
//Set the numFormat for formatting numbers.
  InitLocale();
}

void LocaleEngine::InitLocale()
{
  char *buf=new char[buf_size];
//NumDigits
  memset(buf, 0, buf_size);
  GetLocaleInfo(LOCALE_USER_DEFAULT, LOCALE_IDIGITS, buf, buf_size-1);
  numFormat.NumDigits=atoi(buf);

//LeadingZero
  memset(buf, 0, buf_size);
  GetLocaleInfo(LOCALE_USER_DEFAULT, LOCALE_ILZERO, buf, buf_size-1);
  numFormat.LeadingZero=atoi(buf);

//Grouping
  numFormat.Grouping=3;

//lpDecimalSep
  memset(buf, 0, buf_size);
  GetLocaleInfo(LOCALE_USER_DEFAULT, LOCALE_SDECIMAL, buf, buf_size-1);
  strcpy(numFormat.lpDecimalSep, buf);

//lpThousandSep
  memset(buf, 0, buf_size);
  GetLocaleInfo(LOCALE_USER_DEFAULT, LOCALE_STHOUSAND, buf, buf_size-1);
  strcpy(numFormat.lpThousandSep, buf);

//NegativeOrder
  memset(buf, 0, buf_size);
  GetLocaleInfo(LOCALE_USER_DEFAULT, LOCALE_INEGNUMBER, buf, buf_size-1);
  numFormat.NegativeOrder=atoi(buf);

  delete[] buf;
}

//Format the the number(localize and add thousand separators).
//Remember to make sure cchNumber is 1 less than the size of the lpNumberStr. That is to put a '\0' at the end of the string.
int LocaleEngine::FormatNumberStr(LPCTSTR lpValue, LPTSTR lpNumberStr, int cchNumber)
{
  int n;
  int numDigs;
  numDigs=numFormat.NumDigits;
//keep the decimal digits when formatting the number
  numFormat.NumDigits=DecimalDigits(lpValue);
  n=GetNumberFormat(LOCALE_USER_DEFAULT, 0, lpValue, &numFormat, lpNumberStr, cchNumber);
  numFormat.NumDigits=numDigs;
  return n;
}

//how many digits after the decimal point "."
int LocaleEngine::DecimalDigits(LPCTSTR lpValue)
{
  int n=0;
  const char *p=strrchr(lpValue, '.');
  if (p) {
    n=strlen(lpValue) - (p- lpValue) -1;
    if (n<0)
      n=0;
  }
  return n;
}